bridge              = (require "./../shared/Bridge.coffee").bridge
{ Constants }       = require("./../shared/Constants.coffee")

class exports.PropertiesPanel extends window.FramerStudioComponents.PropertiesPanel

    constructor: (options = {}) ->
        super _.defaults options, {}

        bridge.responders.data = @setData
        bridge.responders.getToggleStates = => @getToggleStates()

        bridge.on "show", => @visible = true
        bridge.on "hide", => @visible = false
        bridge.on Constants.messages.framerview.layerEdited, @handleLayerEdited
        bridge.on "externalValueUpdate", @handleExternalValueUpdate
        bridge.on "undoValue", @handleUndoValue

        @directManipulationProperties =
            x:              round: true, default: "0"
            y:              round: true, default: "0"
            width:          round: true, default: Framer.Defaults.Layer.width.toString()
            height:         round: true, default: Framer.Defaults.Layer.height.toString()
            rotation:       round: true, default: "0"
            rotationX:      round: true, default: "0"
            rotationY:      round: true, default: "0"
            rotationZ:      round: true, default: "0"
            scale:          round: false, default: "1"
            scaleX:         round: false, default: "1"
            scaleY:         round: false, default: "1"
            originX:        round: false, default: "0.5"
            originY:        round: false, default: "0.5"
            originZ:        round: false, default: "0"
            borderRadius:   round: true, default: "0"

        @omitDefaults = true
        @visible = false

    destroy: ->
        @reset()

    setData: (data) =>

        @reset()
        @configure(data)

        @initialValues = data?.layerProperties
        @initialValues = {} unless @initialValues?
        
        for key, value of @directManipulationProperties

            @initialValues[key] = value.default unless @initialValues[key]?

            value.initial = @initialValues[key]
            value.changed = false

        settings = {}
        data.frozenMembers?.forEach (frozenMember) -> settings[frozenMember] = disabled: true

        @setMembers = data.setMembers
        @read(@setMembers, @initialValues, settings)

        @addPropertyListeners()

        window.scrollTo(0, 0);

    getData: () ->
        return @write()

    addPropertyListeners: ->
        for property in @properties
            property.on "change:value", @handlePropertyValueChange
            property.on "clickedOnLock", @handleLockPressed

    removePropertyListeners: ->
        for property in @properties
            property.off "change:value", @handlePropertyValueChange
            property.reset()

    reset: ->
        @removePropertyListeners()

    propertyByDataPath: (dataPath) ->
        for property in @properties
            if property.dataPath is dataPath
                return property

        return null

    handleLayerEdited: (data) =>

        @inBatchMode = true

        changedProperties = []
        for dataPath, settings of @directManipulationProperties

            newValue = data.properties[dataPath]
            property = @propertyByDataPath(dataPath)

            continue if not newValue?
            continue if (settings.initial + "") is (newValue + "") unless property?.changed or @setMembers[dataPath]?

            if settings.round
                value = Math.round(newValue)
            else
                value = newValue

            value = value + "" if value?

            if @omitDefaults
                value = null if "" + value is settings.default

            if value isnt undefined and property?

                if property?

                    if not @omitDefaults and not property.changed and value is property.default + ""
                        value = null

                    currentValue = property.valueToData()
                    if ((value is null and currentValue isnt null) or
                        (value isnt null and value.toString() isnt currentValue)
                    )
                        property.valueFromData(value)
                        property.changed = true
                        changedProperties.push
                            name: dataPath
                            value: value
                            type: property.type

        if changedProperties.length > 0
            bridge.send(
                Constants.messages.propertiespanel.propertiesChanged,
                properties: changedProperties
            )

        @inBatchMode = false

    handlePropertyValueChange: (newValue, oldValue, prop) =>
        return if @inBatchMode

        value = prop.valueToData()
        value = @initialValues[prop.dataPath] unless value?

        update =
            name: prop.dataPath
            value: value
            type: prop.type

        bridge.send(
            Constants.messages.propertiespanel.propertyChanged
            update
        )

    handleLockPressed: (prop) =>
        
        update =
            name: prop.dataPath

        bridge.send(
            Constants.messages.propertiespanel.propertyLockPressed
            update
        )

    # override
    handlePropertyRequest: (request) =>
        switch request?.type
            when "color" then bridge.send "colorPickerRequested", request
            when "image" then bridge.send "imagePickerRequested", request
            when "drag:start"
                document.body.classList.add "dragging"
                bridge.send "drag:start", {}
            when "drag:end"
                document.body.classList.remove "dragging"
                bridge.send "drag:end", {}
            when "indicate" then bridge.send "indicate", request

    handleExternalValueUpdate: (update) =>

        property = @propertyByDataPath(update?.name)
        if property?
            property.value = update.value

    handleUndoValue: (update) =>

        inBatchMode = @inBatchMode
        @inBatchMode = true

        @handleExternalValueUpdate(update)

        @inBatchMode = inBatchMode